#!/usr/bin/python

#-------------------------------------------------------------------------------
#
#   Copyright (C) 2017 Cumulus Networks, Inc. All rights reserved
#
#   This software is subject to the Cumulus Networks End User License
#   Agreement available  at the following locations:
#
#   Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#   Cumulus Linux systems: /usr/share/cumulus/EULA.txt
#
#-------------------------------------------------------------------------------


"""
"""

#-------------------------------------------------------------------------------
#
#   Import the necessary modules
#
#-------------------------------------------------------------------------------

try:
    import argparse
    import chassisd.chassisd_bcm53xx as bcm
    import logging
    import time
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

#-------------------------------------------------------------------------------
#
#   Global Variables
#
#-------------------------------------------------------------------------------

# The handle used for logging messages
logger = logging.getLogger("cumulus-chassisd")


#-------------------------------------------------------------------------------
#
#   The chassisd_accton Class
#
#-------------------------------------------------------------------------------

class chassisd_accton:
    """
    This class implements the functionality of Accton's chassisd daemon, but
    in python.
    """

    #
    # Constants
    #
    _CPLD1_SYSFS   = "/sys/bus/i2c/devices/12-0060/"
    _BCM53XX_SYSFS = "/sys/bus/platform/devices/bcm53xx.0/"

    #
    #   Code
    #
    def __init__(self, chassis, cmdline, stopEvent):
        # Save some command line parameters which we will use later
        self.chassis        = chassis
        self.cmdargs, self.cmdline = self.ParseCmdLine(cmdline)
        self.stopEvent      = stopEvent
        self.bcmPorts       = {}
        self.SetBcmPorts()
        self.SetDefaultTopology()

    def __del__(self):
        self.ClearDefaultTopology()

    def ParseCmdLine(self, cmdline):
        _DEFAULT_LC_MCAST           = "01:10:b5:00:00:01"
        _DEFAULT_FC_MCAST           = "01:10:b5:00:00:02"
        _DEFAULT_STATIC_TOPO        = True
        _DEFAULT_BACKPLANE          = 1

        parser = argparse.ArgumentParser(description="The Accton Chassid Daemon")
        parser.add_argument("--chasLcMcast", default=_DEFAULT_LC_MCAST,
                            metavar="MACADDR", help="The multicast address line cards use.")
        parser.add_argument("--chasFcMcast", default=_DEFAULT_FC_MCAST,
                            metavar="MACADDR", help="The multicast address fabric cards use.")
        parser.add_argument("--chasStaticTopo", action="store_true", default=_DEFAULT_STATIC_TOPO,
                            help="Use a static mgmt network topology")
        parser.add_argument("--chasBackplane", type=self.intBase, default=_DEFAULT_BACKPLANE,
                            metavar="BACKPLANE_VERSION", help="The version of the chassis backplane.")
        args, cmdline = parser.parse_known_args(cmdline)
        return(args, cmdline)

    def intBase(self, intStr):
        '''
        Allow an integer to be specified with standard base representations:
        37 = decimal, 0x25 = hex, 045 = octal, 0b100101 = binary
        '''
        try:
            intVal = int(intStr,0)
        except:
            msg = "%s is not a valid integer" % (intStr,)
            raise argparse.ArgumentTypeError(msg)
        return intVal

    def SignalThreads(self):
        pass

    def JoinThreads(self):
        pass

    def isCpuA(self):
        """
        Returns true if running on CPU A, and false if not
        """
        return self.chassis.GetSlotNum() % 2 == 1

    def isLowerChassis(self):
        """
        Returns true if running on the lower chassis
        """
        slot = self.chassis.GetSlotNum()
        return 17 <= slot <= 32 or 41 <= slot <= 48

    def isFirstLcOrFc(self):
        """
        Returns true if running on CPUA of the first fabric card or line card in
        the uppoer or lower chassis.
        """
        return self.chassis.GetSlotNum() in [1, 17, 33, 41]

    def WriteSysfsNode(self, node, value):
        """
        Writes a sysfs node
        """
        logger.debug("Chassisd - Writing sysfs node %s with value %s" % (node, str(value)))
        try:
            with open(node, 'w') as f:
                f.write(str(value))
        except IOError as e:
            logger.error("Chassisd - I/O error writing %s to %s: %s" %
                         (value, node, e))

    def WriteBcm53xx(self, node, value):
        """
        Write to the Bcm53xx device through the sysfs
        """
        self.WriteSysfsNode(chassisd_accton._BCM53XX_SYSFS + node, value)

    def WriteCpld1(self, node, value):
        """
        Write to the CPLD1 device through the sysfs
        """
        self.WriteSysfsNode(chassisd_accton._CPLD1_SYSFS + node, value)

    def SetBcmPorts(self):
        """
        Fill out the bcmPorts dictionary for the type of card/system we are
        running upon.
        """
        if self.chassis.isFabricCard():
            # First two fabric cards on upper or lower chassis?
            if self.chassis.GetSlotNum() / 4 % 2 == 0:
                if self.cmdargs.chasBackplane == 1:
                    self.bcmPorts = { "CPUA" : 14, "CPUB" : 15, "FC"   : 11, "RJ45" : 9,
                                      "LC0"  : 5,  "LC1"  : 4,  "LC2"  : 7,  "LC3"  : 6,
                                      "LC4"  : 1,  "LC5"  : 0,  "LC6"  : 3,  "LC7"  : 2 }
                else:
                    self.bcmPorts = { "CPUA" : 14, "CPUB" : 15, "FC"   : 11, "RJ45" : 9,
                                      "LC0"  : 4,  "LC1"  : 5,  "LC2"  : 6,  "LC3"  : 7,
                                      "LC4"  : 0,  "LC5"  : 1,  "LC6"  : 2,  "LC7"  : 3 }
            else:
                if self.cmdargs.chasBackplane == 1:
                    self.bcmPorts = { "CPUA" : 14, "CPUB" : 15, "FC"   : 11, "RJ45" : 9,
                                      "LC0"  : 1,  "LC1"  : 0,  "LC2"  : 3,  "LC3"  : 2,
                                      "LC4"  : 5,  "LC5"  : 4,  "LC6"  : 7,  "LC7"  : 6 }
                else:
                    self.bcmPorts = { "CPUA" : 14, "CPUB" : 15, "FC"   : 11, "RJ45" : 9,
                                      "LC0"  : 0,  "LC1"  : 1,  "LC2"  : 2,  "LC3"  : 3,
                                      "LC4"  : 4,  "LC5"  : 5,  "LC6"  : 6,  "LC7"  : 7 }
        else:
            self.bcmPorts = { "CPUA" : 0,  "CPUB" : 7,
                              "FC0"  : 2,  "FC1"  : 3,  "FC2"  : 4,  "FC3"  : 5 }

    def ResetBcm53xx(self):
        """
        Reset the Bcm53xx chip via the reset line on the CPLD
        """
        self.WriteCpld1("reset_bcm53xx", 1)
        time.sleep(0.020)
        self.WriteCpld1("reset_bcm53xx", 0)

    def SetInitialPhy(self):
        """
        Set the initial state of the PHY. First set the default value in the
        BMC53xx external PHY scan control register, then reset the phy, and
        finally set the autonegotation advertisement.
        """
        self.WriteBcm53xx("extPhyCntrl", bcm._EXT_PHY_CTRL_ADDR_INC)
        self.WriteCpld1("reset_mgmt_phy", 1)
        time.sleep(0.001)
        self.WriteCpld1("reset_mgmt_phy", 0)
        self.WriteBcm53xx("port10AutoNegAdv",
                           bcm._PORT_AUTO_NEG_ADVER_100BASET4 |\
                           bcm._PORT_AUTO_NEG_ADVER_100FD     |\
                           bcm._PORT_AUTO_NEG_ADVER_100HD     |\
                           bcm._PORT_AUTO_NEG_ADVER_10FD      |\
                           bcm._PORT_AUTO_NEG_ADVER_10HD      |\
                           bcm._PORT_AUTO_NEG_ADVER_SEL_8023)

    def SetSGMIIMode(self, mode):
        """
        Configure the SGMII interfaces on a card
        """
        for port in self.bcmPorts:
            if port not in set(["CPUA", "CPUB"]):
                self.WriteBcm53xx("port%dSGMIICntrl" % (self.bcmPorts[port],), mode)

    def SetStaticARL(self):
        """
        Configure the multicast address used for card-card communication
        """
        mcAddr = self.cmdargs.chasFcMcast if self.chassis.isFabricCard() else self.cmdargs.chasLcMcast
        if self.isLowerChassis():
            mcAddr = self.cmdargs.chasLcMcast if self.chassis.isFabricCard() else self.cmdargs.chasFcMcast
        mcVect = 1 << self.bcmPorts["CPUA"]

        self.WriteBcm53xx("globalArlCfg", bcm._GLOBAL_ARL_CFG_MPORT_ADDR_EN)
        self.WriteBcm53xx("bpduMcastAddr", mcAddr)
        self.WriteBcm53xx("multiportAddr1", mcAddr)
        self.WriteBcm53xx("multiportVec1", mcVect)

    def ConfigPortStp(self, port, state):
        """
        Configure the STP state of a port
        """
        self.WriteBcm53xx("port%dCntrl" % (port), state)

    def SetDefaultTopology(self):
        """
        Programs the BMC53xx with the default topology.
        """
        # The BCM53xx can only be accessed through CPU A
        if self.isCpuA():
            self.ResetBcm53xx()
            if self.chassis.isFabricCard():
                # Fabric card initialization
                self.SetInitialPhy()
                modeReg = "switchMode96"
                sgmiiMode = bcm._PORT_SGMII_CNTRL_CD_EN      |\
                            bcm._PORT_SGMII_CNTRL_CRC_DIS    |\
                            bcm._PORT_SGMII_CNTRL_RSVD       |\
                            bcm._PORT_SGMII_CNTRL_AUTODET_EN |\
                            bcm._PORT_SGMII_CNTRL_FIBER
            else:
                # Line card initialization
                modeReg = "switchMode89"
                sgmiiMode = bcm._PORT_SGMII_CNTRL_CD_EN      |\
                            bcm._PORT_SGMII_CNTRL_CRC_DIS    |\
                            bcm._PORT_SGMII_CNTRL_RSVD       |\
                            bcm._PORT_SGMII_CNTRL_FIBER

            self.SetSGMIIMode(sgmiiMode)
            self.WriteBcm53xx(modeReg,
                              bcm._SWITCH_MODE_RTRY_LMT_DIS |
                              bcm._SWITCH_MODE_SW_FWDG_EN   |
                              bcm._SWITCH_MODE_SW_FWDG_MODE)
            self.SetStaticARL()
            for port in self.bcmPorts:
                state = bcm._PORT_CTRL_STP_STATE_BLOCKING
                if port in set(["CPUA", "CPUB", "RJ45"]):
                    state = bcm._PORT_CTRL_STP_STATE_FORWARDING
                if self.cmdargs.chasStaticTopo:
                    if port in set(["LC0", "FC0"]) or self.isFirstLcOrFc():
                        state = bcm._PORT_CTRL_STP_STATE_FORWARDING
                self.ConfigPortStp(self.bcmPorts[port], state)

    def ClearDefaultTopology(self):
        """
        Breaks down the default topology so that loops can't form.
        """
        if self.isCpuA() and not self.cmdargs.chasStaticTopo:
            for port in self.bcmPorts:
                state = bcm._PORT_CTRL_STP_STATE_BLOCKING
                if port in set(["CPUA", "CPUB", "RJ45"]):
                    state = bcm._PORT_CTRL_STP_STATE_FORWARDING
                self.ConfigPortStp(self.bcmPorts[port], state)


if __name__ == '__main__':
    pass
